from datetime import datetime
import argparse
import requests
import numpy as np
from keras.callbacks import Callback
from slack_sdk import WebClient
from slackker.utils import checkker
from slackker.utils import functions
from slackker.utils.ccolors import colors

class SlackUpdate(Callback):
    """Custom Keras callback that posts to Slack while training a neural network"""
    def __init__(self, token, channel, ModelName, export="png", SendPlot=False, verbose=0):

        if token is None:
            colors.prRed('[slackker] ERROR: Please enter Valid Slack API Token.')
            return

        server = checkker.check_internet(url="www.slack.com", verbose=verbose)
        api = checkker.slack_connect(token=token, verbose=verbose)

        if server and api:
            self.client = WebClient(token=token)
            self.channel = channel
            self.ModelName = ModelName
            self.export = export
            self.SendPlot = SendPlot
            self.verbose = verbose

            if export is not None:
                pass
            else:
                raise argparse.ArgumentTypeError("[slackker] ERROR: 'export' argument is missing (supported formats: eps, jpeg, jpg, pdf, pgf, png, ps, raw, rgba, svg, svgz, tif, tiff)")

        self.train_acc = []
        self.valid_acc = []
        self.train_loss = []
        self.valid_loss = []
        self.n_epochs = 0

    # Called when training starts
    def on_train_begin(self, logs={}):
        functions.Slack.report_stats(
            client=self.client,
            channel=self.channel,
            text=f'Training on "{self.ModelName}" started at {datetime.now().strftime("%Y-%m-%d %H:%M:%S")}',
            verbose=self.verbose)

    # Called when epoch ends
    def on_epoch_end(self, batch, logs={}):

        custom_logs = []

        for i in logs:
            custom_logs.append(logs[i])

        self.train_loss.append(custom_logs[0])
        self.train_acc.append(custom_logs[1])
        self.valid_loss.append(custom_logs[2])
        self.valid_acc.append(custom_logs[3])

        message = f'Epoch: {self.n_epochs}, Training Loss: {self.train_loss[-1]:.4f}, Validation Loss: {self.valid_loss[-1]:.4f}'

        # Check internet before sending update on slacj
        server = checkker.check_internet_epoch_end(url="www.slack.com")

        # If internet working send message else skip sending message and continue training.
        if server:
            functions.Slack.report_stats(
                client=self.client,
                channel=self.channel,
                text=message,
                verbose=self.verbose)
        else:
            pass

        self.n_epochs += 1

    # Prepare and send report with graphs at the end of training.
    def on_train_end(self, logs={}):

        best_epoch = np.argmin(self.valid_loss)
        val_loss = self.valid_loss[best_epoch]
        train_loss = self.train_loss[best_epoch]
        train_acc = self.train_acc[best_epoch]
        val_acc = self.valid_acc[best_epoch]

        training_logs = {'train_loss': self.train_loss, 'train_acc': self.train_acc, 'val_loss': self.valid_loss, 'val_acc': self.valid_acc}

        message1 = f'Trained for {self.n_epochs} epochs. Best epoch was {best_epoch}.'
        message2 = f"Best validation loss = {val_loss:.4f}, Training Loss = {train_loss:.4f}, Best Accuracy = {100*val_acc:.4f}%"

        functions.Slack.report_stats(client=self.client, channel=self.channel, text=message1, verbose=self.verbose)

        functions.Slack.report_stats(client=self.client, channel=self.channel, text=message2, verbose=self.verbose)

        functions.Slack.keras_plot_history(ModelName=self.ModelName,
            export=self.export,
            client=self.client,
            channel=self.channel,
            SendPlot=self.SendPlot,
            training_logs=training_logs,
            verbose=self.verbose)

class TelegramUpdate(Callback):
    """Custom Keras callback that posts to Telegram while training a neural network"""
    def __init__(self, token, ModelName, export="png", SendPlot=False, verbose=0):

        if token is None:
            colors.prRed('[slackker] ERROR: Please enter Valid Telegram API Token.')
            return

        server = checkker.check_internet(url="www.telegram.org", verbose=verbose)
        channel = checkker.get_telegram_chat_id(token=token, verbose=verbose)

        if server and channel:
            self.token = token
            self.channel = channel
            self.ModelName = ModelName
            self.export = export
            self.SendPlot = SendPlot
            self.verbose = verbose

            if export is not None:
                pass
            else:
                raise argparse.ArgumentTypeError("[slackker] ERROR: 'export' argument is missing (supported formats: eps, jpeg, jpg, pdf, pgf, png, ps, raw, rgba, svg, svgz, tif, tiff)")

        self.train_acc = []
        self.valid_acc = []
        self.train_loss = []
        self.valid_loss = []
        self.n_epochs = 0

    # Called when training starts
    def on_train_begin(self, logs={}):
        functions.Telegram.report_stats(
            token=self.token,
            channel=self.channel,
            text=f'Training on "{self.ModelName}" started at {datetime.now().strftime("%Y-%m-%d %H:%M:%S")}',
            verbose=self.verbose)

    # Called when epoch ends
    def on_epoch_end(self, batch, logs={}):

        custom_logs = []

        for i in logs:
            custom_logs.append(logs[i])

        self.train_loss.append(custom_logs[0])
        self.train_acc.append(custom_logs[1])
        self.valid_loss.append(custom_logs[2])
        self.valid_acc.append(custom_logs[3])
        self.n_epochs += 1

        message = f'Epoch: {self.n_epochs}, Training Loss: {self.train_loss[-1]:.4f}, Validation Loss: {self.valid_loss[-1]:.4f}'

        # Check internet before sending update on slacj
        server = checkker.check_internet_epoch_end(url="www.telegram.org")

        # If internet working send message else skip sending message and continue training.
        if server:
            functions.Telegram.report_stats(
                token=self.token,
                channel=self.channel,
                text=message,
                verbose=self.verbose)
        else:
            pass

    # Prepare and send report with graphs at the end of training.
    def on_train_end(self, logs={}):

        best_epoch = np.argmin(self.valid_loss)
        val_loss = self.valid_loss[best_epoch]
        train_loss = self.train_loss[best_epoch]
        train_acc = self.train_acc[best_epoch]
        val_acc = self.valid_acc[best_epoch]

        training_logs = {'train_loss': self.train_loss, 'train_acc': self.train_acc, 'val_loss': self.valid_loss, 'val_acc': self.valid_acc}

        message1 = f'Trained for {self.n_epochs} epochs. Best epoch was {best_epoch}.'
        message2 = f"Best validation loss = {val_loss:.4f}, Training Loss = {train_loss:.4f}, Best Accuracy = {100*val_acc:.4f}%"

        functions.Telegram.report_stats(token=self.token, channel=self.channel, text=message1, verbose=self.verbose)

        functions.Telegram.report_stats(token=self.token, channel=self.channel, text=message2, verbose=self.verbose)

        functions.Telegram.keras_plot_history(ModelName=self.ModelName,
            export=self.export,
            token=self.token,
            channel=self.channel,
            SendPlot=self.SendPlot,
            training_logs=training_logs,
            verbose=self.verbose)


class colors:
	# Assign colors to logging statements
	def prRed(sent):
		print("\033[91m {}\033[00m" .format(sent))

	def prCyan(sent):
		print("\033[96m {}\033[00m" .format(sent))

	def prYellow(sent):
		print("\033[93m {}\033[00m" .format(sent))


from slackker import callbacks
from slackker.utils import checkker
from slackker.utils import functions

__author__ = 'Siddhesh Gunjal'
__email__ = 'siddhu19@live.com'
__version__ = '1.2.33'

# module level doc-string
__doc__ = """
slackker
=====================================================================
Description
-----------
slackker is a python package for monitoring your python script & Model training status in real-time on slack & telegram.

References
----------
https://github.com/siddheshgunjal/slackker
"""


import requests
import itertools
import matplotlib.pyplot as plt
from slack_sdk.errors import SlackApiError
from slackker.utils.ccolors import colors

class Plotter:
    @staticmethod
    def slack_img_upload(name, client, channel, filepath, verbose=1):
        """Upload generated graphs"""
        try:
            # Call the chat.postMessage method using the WebClient
            response = client.files_upload_v2(channel=channel,
            file = filepath,
            initial_comment=f"{name} :bar_chart:")
            if verbose>=1:
                colors.prCyan(f"[slackker] DEBUG: Uploaded graph on {channel} channel")

        except SlackApiError as e:
            colors.prRed(f"[slackker] ERROR: Error uploading graph: {e}")
    
    @staticmethod
    def telegram_img_upload(name, token, channel, image, verbose=1):
        # apiURL for send image
        apiURL = f'https://api.telegram.org/bot{token}/sendPhoto'

        try:
            # Call the requests.post method using API request
            response = requests.post(apiURL, params={'chat_id': channel, 'caption': f"{name} \U0001F4CA"}, files={'photo': image})
            if verbose>=1:
                colors.prCyan("[slackker] DEBUG: Uploaded graph on Telegram")

        except Exception as e:
            colors.prRed(f"[slackker] ERROR: Error uploading graph: {e}")

    @staticmethod
    def plot_and_upload(platform, ModelName, export, client, channel, SendPlot, logs, metric, verbose=1):
        try:
            # Make sure training has began
            if len(logs) == 0:
                colors.prRed('[slackker] ERROR: Loss is missing from training history')
                return

            # As loss always exists
            k = str(list(logs.keys())[0])
            epochs = range(0, len(logs[k]))

            # color choices list
            clrs = ['b-', 'g-', 'r-', 'c-', 'm-', 'y-', 'k-', 'orange', 'darkviolet', 'fuchsia']

            # seperate color iterator for loss and acc graphs
            ls, acc = itertools.tee(clrs, 2) # create as many as needed

            path = f'{ModelName}_{metric}.{export}'

            # Plotting
            plt.figure(1, figsize=(15,8))

            for key, value in logs.items():
                if metric in key.lower():
                    plt.plot(epochs, logs[key], next(ls) if metric == "loss" else next(acc), lw=2.5, label=f'{key}: {logs[key][-1]:.4f}')
                else:
                    pass
            plt.title(f'{ModelName}_{metric}_Graph', fontsize=20)
            plt.xlabel('Epochs', fontsize=15)
            plt.ylabel(f'{metric}', fontsize=15)
            plt.legend(fontsize=12)
            plt.grid(True)
            plt.savefig(path)
            plt.close()

            if SendPlot:
                try:
                    if platform == 'slack':
                        Plotter.slack_img_upload(name = f'{ModelName}_{metric}', client=client, channel=channel, filepath=path, verbose=verbose)
                    else:
                        Plotter.telegram_img_upload(name = f'{ModelName}_{metric}', token=client, channel=channel, image=open(path, 'rb'), verbose=verbose)
                except Exception as e:
                    colors.prRed(f"[slackker] ERROR: Invalid Argument: {e}")
            else:
                colors.prYellow("[slackker] WARNING: Skipping graph upload as SendPlot == False ")
        except Exception as e:
            colors.prRed(f'[slackker] ERROR: Plotting Generation Failed: {e}')

class Slack:
    @staticmethod
    def report_stats(client, channel, text, verbose=1):
        """Report training stats"""
        try:
            # Call the chat.postMessage method using the WebClient
            result = client.chat_postMessage(
                channel=channel, 
                text=text)
            if verbose>=1:
                colors.prCyan(f"[slackker] Posted update on {channel} channel")

        except SlackApiError as e:
            colors.prRed(f"[slackker] ERROR: Error posting update: {e}")

    @staticmethod
    def keras_plot_history(ModelName, export, client, channel, SendPlot, training_logs, verbose=1):
        #plot loss
        Plotter.plot_and_upload(platform='slack', ModelName=ModelName, export=export, client=client, channel=channel, SendPlot=SendPlot, logs=training_logs, metric="loss", verbose=verbose)

        #plot accuracy
        Plotter.plot_and_upload(platform='slack', ModelName=ModelName, export=export, client=client, channel=channel, SendPlot=SendPlot, logs=training_logs, metric="acc", verbose=verbose)

    @staticmethod
    def lightning_plot_history(ModelName, export, client, channel, SendPlot, training_logs, verbose=1):
        #plot loss
        Plotter.plot_and_upload(platform='slack', ModelName=ModelName, export=export, client=client, channel=channel, SendPlot=SendPlot, logs=training_logs, metric="loss", verbose=verbose)

        #plot accuracy
        Plotter.plot_and_upload(platform='slack', ModelName=ModelName, export=export, client=client, channel=channel, SendPlot=SendPlot, logs=training_logs, metric="acc", verbose=verbose)

class Telegram:
    @staticmethod
    def report_stats(token, channel, text, verbose=1):
        # apiURL for send message
        apiURL = f'https://api.telegram.org/bot{token}/sendMessage'

        try:
            # Call the requests.post method using API request
            response = requests.post(apiURL, params={'chat_id': channel, 'text': text})
            if verbose>=1:
                colors.prCyan("[slackker] DEBUG: Posted update on Telegram")

        except Exception as e:
            colors.prRed(f"[slackker] ERROR: Error posting update: {e}")

    @staticmethod
    def keras_plot_history(ModelName, export, token, channel, SendPlot, training_logs, verbose=1):
        #plot loss
        Plotter.plot_and_upload(platform='telegram', ModelName=ModelName, export=export, client=token, channel=channel, SendPlot=SendPlot, logs=training_logs, metric="loss", verbose=verbose)

        #plot accuracy
        Plotter.plot_and_upload(platform='telegram', ModelName=ModelName, export=export, client=token, channel=channel, SendPlot=SendPlot, logs=training_logs, metric="acc", verbose=verbose)

    @staticmethod
    def lightning_plot_history(ModelName, export, token, channel, SendPlot, training_logs, verbose=1):
        #plot loss
        Plotter.plot_and_upload(platform='telegram', ModelName=ModelName, export=export, client=token, channel=channel, SendPlot=SendPlot, logs=training_logs, metric="loss", verbose=verbose)

        #plot accuracy
        Plotter.plot_and_upload(platform='telegram', ModelName=ModelName, export=export, client=token, channel=channel, SendPlot=SendPlot, logs=training_logs, metric="acc", verbose=verbose)


import time
import requests
import http.client as httplib
from slack_sdk import WebClient
from datetime import datetime
from slackker.utils.ccolors import colors

# Check whether server is still alive.
def check_internet(url, verbose=2):
    counter=1
    status=False
    sleepinsec=60

    while status is False:
        conn = httplib.HTTPConnection(url, timeout=5)
        try:
            conn.request("HEAD", "/")
            conn.close()
            status=True
            if verbose >= 2:
                colors.prCyan(f"[slackker] DEBUG: {datetime.now().strftime('%d-%m-%Y %H:%M:%S')} Connection to '{url}' server successful!")
        except Exception:
            status=False
            colors.prYellow(f"[slackker] WARNING:: {datetime.now().strftime('%d-%m-%Y %H:%M:%S')} Connection to '{url}' server failed. Please check your internet. Trying again in 60 sec..[attempt {counter}]")
            time.sleep(sleepinsec)
            counter=counter+1

    if counter>1:
        if verbose>=2:
            colors.prCyan(f"[slackker] DEBUG: {datetime.now().strftime('%d-%m-%Y %H:%M')} re-established connection to '{url}' server after {counter} attempts.")

    return status

# Check whether server is still alive at the end of epoch if not skip the training. 
def check_internet_epoch_end(url):
    counter=1
    status=False
    sleepinsec=10

    while status is False and counter <= 3:
        conn = httplib.HTTPConnection(url, timeout=5)
        try:
            conn.request("HEAD", "/")
            conn.close()
            status=True
        except Exception:
            status=False
            colors.prYellow(f"[slackker] WARNING: {datetime.now().strftime('%d-%m-%Y %H:%M:%S')} Connection to '{url}' server failed. Trying again in 10 sec..[attempt {counter}]")
            time.sleep(sleepinsec)
            counter=counter+1
                
    if counter <= 3 and counter > 1:
        colors.prCyan(f"[slackker] DEBUG: {datetime.now().strftime('%d-%m-%Y %H:%M:%S')} re-established connection to '{url}' server after {counter} attempts.")
    elif counter > 3:
        colors.prYellow(f'[slackker] WARNING: Skipping report update due to connection failure. {counter} attempts made before skipping')

    return status


def slack_connect(token, verbose=2):
    status = False

    try:
        client = WebClient(token=token)
        api_response = client.api_test()
        status = True
        if verbose >= 2:
            colors.prCyan(f"[slackker] DEBUG: Connection to slack API successful! {api_response}")
    except Exception as e:
        status = False
        colors.prRed(f"[slackker] ERROR: Invalid slack API token: {e}")

    return status

def get_telegram_chat_id(token, verbose=2):
    chat_id = False
    url = f"https://api.telegram.org/bot{token}/getUpdates"

    try:
        firstMsg = requests.get(url).json()
        chat_id = str(firstMsg["result"][0]["message"]["chat"]["id"])
        if verbose >= 2:
            colors.prCyan(f"[slackker] DEBUG: Connection to telegram API successful!")
            colors.prCyan(f"[slackker] DEBUG: Found chat with 'chat_id'={chat_id}")
    except Exception as e:
        chat_id = False
        colors.prRed(f"[slackker] ERROR: Could not connect to Telegram API: {e}")
        colors.prYellow(f"[slackker] SUGGESTION: Please send 'Hello' once to your bot to make it discoverable to slackker")

    return chat_id



import time
import os
from inspect import stack
from datetime import datetime
from slack_sdk import WebClient
from slackker.utils import checkker
from slackker.utils import functions
from slackker.utils.ccolors import colors

class SlackUpdate():
    ''' SlackUpdate class to send updates to Telegram channel '''
    def __init__(self, token, channel, verbose=0):

        if token is None:
            colors.prRed('[slackker] ERROR: Please enter Valid Slack API Token.')
            return

        server = checkker.check_internet(url="www.telegram.org", verbose=verbose)
        api = checkker.slack_connect(token=token, verbose=verbose)

        if server and api:
            self.client = WebClient(token=token)
            self.channel = channel
            self.verbose = verbose
        
    def notifier(self, function):
        ''' Decorator to log function calls '''
        def wrapper(*args, **kwargs):
            if self.verbose > 0:
                # Log the function call
                colors.prCyan(f"[slackker] INFO: Calling {function.__name__} with args: {args}, kwargs: {kwargs}")
            
            # Call the original function
            start_time = time.time()
            result = function(*args, **kwargs)
            end_time = time.time()

            # execution time
            execution_time = end_time - start_time

            if result is not None:
                if isinstance(result, tuple):
                    message = f"Function '{function.__name__}' from Script: '{os.path.basename(__import__(function.__module__).__file__)}' executed.\nExecution time: {execution_time:.3f} Seconds\nReturned {len(result)} outputs:\n"
                    num = 0
                    for i in result:
                        message += f"Output {num}:\n{i}\n\n"
                        num += 1
                else:
                    message = f"Function '{function.__name__}' from Script: '{os.path.basename(__import__(function.__module__).__file__)}' executed.\nExecution time: {execution_time:.3f} Seconds\nReturned output: {result}"
            else:
                message = f"Function '{function.__name__}' from Script: '{os.path.basename(__import__(function.__module__).__file__)}' executed.\nExecution time: {execution_time:.3f} Seconds\nReturned output: None"

            # Log the return value
            functions.Slack.report_stats(
                client=self.client,
                channel=self.channel,
                text=message,
                verbose=self.verbose
            )

            return result
        return wrapper
    
    def notify(self, *args, **kwargs):
        ''' Notify the user that the script has been executed '''
        
        script = stack()[1].filename
        
        text = f"Your script: '{os.path.basename(script)}' has been executed successfully at {datetime.now().strftime('%d-%m-%Y %H:%M:%S')}"
        if args:
            for arg in args:
                text += f"\n\n{arg}"
        if kwargs:
            for key, value in kwargs.items():
                text += f"\n\n{key}: {value}"

        functions.Slack.report_stats(
            client=self.client,
            channel=self.channel,
            text=text,
            verbose=self.verbose
        )

class TelegramUpdate():
    ''' TelegramUpdate class to send updates to Telegram channel '''
    def __init__(self, token, verbose=0):

        if token is None:
            colors.prRed('[slackker] ERROR: Please enter Valid Telegram API Token.')
            return

        server = checkker.check_internet(url="www.telegram.org", verbose=verbose)
        channel = checkker.get_telegram_chat_id(token=token, verbose=verbose)

        if server and channel:
            self.token = token
            self.channel = channel
            self.verbose = verbose
        
    def notifier(self, function):
        ''' Decorator to log function calls '''
        def wrapper(*args, **kwargs):
            if self.verbose > 0:
                # Log the function call
                colors.prCyan(f"[slackker] INFO: Calling {function.__name__} with args: {args}, kwargs: {kwargs}")
            
            # Call the original function
            start_time = time.time()
            result = function(*args, **kwargs)
            end_time = time.time()

            # execution time
            execution_time = end_time - start_time

            if result is not None:
                if isinstance(result, tuple):
                    message = f"Function '{function.__name__}' from Script: '{os.path.basename(__import__(function.__module__).__file__)}' executed.\nExecution time: {execution_time:.3f} Seconds\nReturned {len(result)} outputs:\n"
                    num = 0
                    for i in result:
                        message += f"Output {num}:\n{i}\n\n"
                        num += 1
                else:
                    message = f"Function '{function.__name__}' from Script: '{os.path.basename(__import__(function.__module__).__file__)}' executed.\nExecution time: {execution_time:.3f} Seconds\nReturned output: {result}"
            else:
                message = f"Function '{function.__name__}' from Script: '{os.path.basename(__import__(function.__module__).__file__)}' executed.\nExecution time: {execution_time:.3f} Seconds\nReturned output: None"

            # Log the return value
            functions.Telegram.report_stats(
                token=self.token,
                channel=self.channel,
                text=message,
                verbose=self.verbose
            )

            return result
        return wrapper
    
    def notify(self, *args, **kwargs):
        ''' Notify the user that the script has been executed '''

        script = stack()[1].filename
        
        text = f"Your script: '{os.path.basename(script)}' has been executed successfully at {datetime.now().strftime('%d-%m-%Y %H:%M:%S')}"
        if args:
            for arg in args:
                text += f"\n\n{arg}"
        if kwargs:
            for key, value in kwargs.items():
                text += f"\n\n{key}: {value}"

        functions.Telegram.report_stats(
            token=self.token,
            channel=self.channel,
            text=text,
            verbose=self.verbose
        )


import setuptools
import re
import os

# versioning ------------
VERSIONFILE="slackker/__init__.py"
getversion = re.search( r"^__version__ = ['\"]([^'\"]*)['\"]", open(VERSIONFILE, "rt").read(), re.M)
if getversion:
    new_version = getversion.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))

# Setup ------------
def read_file(file_name):
    path = os.path.join(os.path.dirname(__file__), file_name)
    with open(path, 'r', encoding='utf-8') as f:
        return f.read()

setuptools.setup(
     install_requires=['slack_sdk>=3.19.0','numpy','matplotlib'],
     python_requires='>=3.5',
     name='slackker',
     version=new_version,
     license="Apache-2.0",
     author="Siddhesh Gunjal",
     author_email="siddhu19@live.com",
     description="Python package for monitoring your python script & Model training status in real-time on slack & telegram.",
     long_description=read_file("README.md"),
     long_description_content_type="text/markdown",
     url="https://github.com/siddheshgunjal/slackker",
     packages=setuptools.find_packages(), # Searches throughout all dirs for files to include
     include_package_data=True, # Must be true to include files depicted in MANIFEST.in
     classifiers=[
         "Programming Language :: Python :: 3",
         "License :: OSI Approved :: Apache Software License",
         "Operating System :: OS Independent",
         "Intended Audience :: Developers",
         "Intended Audience :: Information Technology",
         "Intended Audience :: Science/Research",
     ],
 )


import sys
from datetime import datetime
import argparse
import requests
import numpy as np
from lightning.pytorch.callbacks import Callback
from slack_sdk import WebClient
from slackker.utils import checkker
from slackker.utils import functions
from slackker.utils.ccolors import colors

class SlackUpdate(Callback):
    """Custom Lightning callback that posts to Slack while training a neural network"""
    def __init__(self, token, channel, ModelName, TrackLogs=None, monitor=None, export="png", SendPlot=False, verbose=0):
        if token is None:
            colors.prRed('[slackker] ERROR: Please enter Valid Slack API Token.')
            return

        server = checkker.check_internet(url="www.slack.com", verbose=verbose)
        api = checkker.slack_connect(token=token, verbose=verbose)

        if server and api:
            self.client = WebClient(token=token)
            self.channel = channel
            self.ModelName = ModelName
            self.export = export
            self.SendPlot = SendPlot
            self.verbose = verbose
            self.TrackLogs = TrackLogs
            self.monitor = monitor

            if export is not None:
                pass
            else:
                raise argparse.ArgumentTypeError("[slackker] ERROR: 'export' argument is missing (supported formats: eps, jpeg, jpg, pdf, pgf, png, ps, raw, rgba, svg, svgz, tif, tiff)")

            if TrackLogs is None:
                colors.prRed("[slackker] ERROR: Provice at least 1 log type for sending update.")
                sys.exit()
            else:
                if type(TrackLogs) is not list and TrackLogs is not None:
                    colors.prRed("[slackker] ERROR: 'TrackLogs' is a list type of argument, add values in '[]'")
                    sys.exit()
                else:
                    pass

            if monitor is not None:
                if monitor not in TrackLogs:
                    colors.prRed("[slackker] ERROR: Couldn't find Argument 'monitor' value in 'TrackLogs' provided")
                    sys.exit()
                else:
                    pass
            else:
                colors.prYellow("[slackker] WARNING: Argument 'monitor' not provided, will skip reporting Best Epoch")

        self.training_logs = {}
        self.n_epochs = 0

    # Called when training starts
    def on_fit_start(self, trainer, pl_module):
        functions.Slack.report_stats(
            client=self.client,
            channel=self.channel,
            text=f'Training on "{self.ModelName}" started at {datetime.now().strftime("%Y-%m-%d %H:%M:%S")}',
            verbose=self.verbose)

    # Called when every training epoch ends
    def on_train_epoch_end(self, trainer, pl_module):
        metrics = trainer.callback_metrics
        logs = self.TrackLogs

        custom_logs = {}
        toPrint = []

        [custom_logs.update({i:float(metrics[i])}) for i in logs]

        [self.training_logs.setdefault(key, []).append(value) for key, value in custom_logs.items()]

        [toPrint.append(f"{i}: {metrics[i]:.4f}") for i in logs]

        message = f"Epoch: {self.n_epochs}, {', '.join(toPrint)}"

        # Check internet before sending update on slacj
        server = checkker.check_internet_epoch_end(url="www.slack.com")

        # If internet working send message else skip sending message and continue training.
        if server:
            functions.Slack.report_stats(
                client=self.client,
                channel=self.channel,
                text=message,
                verbose=self.verbose)
        else:
            pass

        self.n_epochs += 1

    # Prepare and send report with graphs at the end of training.
    def on_fit_end(self, trainer, pl_module):
        if self.monitor is not None:
            for key, value in self.training_logs.items():
                if "loss" in self.monitor.lower():
                    if self.monitor.lower() in key.lower():
                        # print(f"Lowest loss value is {min(value)}")
                        message = f"Trained for {self.n_epochs} epochs. Best epoch was, Epoch {np.argmin(value)}"
                elif "acc" in self.monitor.lower():
                    if self.monitor.lower() in key.lower():
                        # print(f"Highest accuracy value is{max(value)}")
                        message = f"Trained for {self.n_epochs} epochs. Best epoch was, Epoch {np.argmax(value)}"
        else:
            message = f"Trained for {self.n_epochs} epochs. Argument 'monitor' was not provided, skipped reporting Best Epoch"

        functions.Slack.report_stats(
            client=self.client,
            channel=self.channel,
            text=message,
            verbose=self.verbose)

        functions.Slack.lightning_plot_history(ModelName=self.ModelName,
            export=self.export,
            client=self.client,
            channel=self.channel,
            SendPlot=self.SendPlot,
            training_logs=self.training_logs,
            verbose=self.verbose)

class TelegramUpdate(Callback):
    """Custom Lightning callback that posts to Telegram while training a neural network"""
    def __init__(self, token, ModelName, TrackLogs=None, monitor=None, export="png", SendPlot=False, verbose=0):
        if token is None:
            colors.prRed('[slackker] ERROR: Please enter Valid Telegram API Token.')
            return

        server = checkker.check_internet(url="www.telegram.org", verbose=verbose)
        channel = checkker.get_telegram_chat_id(token=token, verbose=verbose)

        if server and channel:
            self.token = token
            self.channel = channel
            self.ModelName = ModelName
            self.export = export
            self.SendPlot = SendPlot
            self.verbose = verbose
            self.TrackLogs = TrackLogs
            self.monitor = monitor

            if export is not None:
                pass
            else:
                raise argparse.ArgumentTypeError("[slackker] ERROR: 'export' argument is missing (supported formats: eps, jpeg, jpg, pdf, pgf, png, ps, raw, rgba, svg, svgz, tif, tiff)")

            if TrackLogs is None:
                colors.prRed("[slackker] ERROR: Provice at least 1 log type for sending update.")
                sys.exit()
            else:
                if type(TrackLogs) is not list and TrackLogs is not None:
                    colors.prRed("[slackker] ERROR: 'TrackLogs' is a list type of argument, add values in '[]'")
                    sys.exit()
                else:
                    pass

            if monitor is not None:
                if monitor not in TrackLogs:
                    colors.prRed("[slackker] ERROR: Couldn't find Argument 'monitor' value in 'TrackLogs' provided")
                    sys.exit()
                else:
                    pass
            else:
                colors.prYellow("[slackker] WARNING: Argument 'monitor' not provided, will skip reporting Best Epoch")

        self.training_logs = {}
        self.n_epochs = 0

    # Called when training starts
    def on_fit_start(self, trainer, pl_module):
        functions.Telegram.report_stats(
            token=self.token,
            channel=self.channel,
            text=f'Training on "{self.ModelName}" started at {datetime.now().strftime("%Y-%m-%d %H:%M:%S")}',
            verbose=self.verbose)

    # Called when every training epoch ends
    def on_train_epoch_end(self, trainer, pl_module):
        metrics = trainer.callback_metrics
        logs = self.TrackLogs

        custom_logs = {}
        toPrint = []

        [custom_logs.update({i:float(metrics[i])}) for i in logs]

        [self.training_logs.setdefault(key, []).append(value) for key, value in custom_logs.items()]

        [toPrint.append(f"{i}: {metrics[i]:.4f}") for i in logs]

        message = f"Epoch: {self.n_epochs}, {', '.join(toPrint)}"

        # Check internet before sending update on slacj
        server = checkker.check_internet_epoch_end(url="www.slack.com")

        # If internet working send message else skip sending message and continue training.
        if server:
            functions.Telegram.report_stats(
                token=self.token,
                channel=self.channel,
                text=message,
                verbose=self.verbose)
        else:
            pass

        self.n_epochs += 1

    # Prepare and send report with graphs at the end of training.
    def on_fit_end(self, trainer, pl_module):
        if self.monitor is not None:
            for key, value in self.training_logs.items():
                if "loss" in self.monitor.lower():
                    if self.monitor.lower() in key.lower():
                        # print(f"Lowest loss value is {min(value)}")
                        message = f"Trained for {self.n_epochs} epochs. Best epoch was, Epoch {np.argmin(value)}"
                elif "acc" in self.monitor.lower():
                    if self.monitor.lower() in key.lower():
                        # print(f"Highest accuracy value is{max(value)}")
                        message = f"Trained for {self.n_epochs} epochs. Best epoch was, Epoch {np.argmax(value)}"
        else:
            message = f"Trained for {self.n_epochs} epochs. Argument 'monitor' was not provided, skipped reporting Best Epoch"

        functions.Telegram.report_stats(
            token=self.token,
            channel=self.channel,
            text=message,
            verbose=self.verbose)

        functions.Telegram.lightning_plot_history(ModelName=self.ModelName,
            export=self.export,
            token=self.token,
            channel=self.channel,
            SendPlot=self.SendPlot,
            training_logs=self.training_logs,
            verbose=self.verbose)